import warnings
import time
import os
from pathlib import Path

from .load_data import load_data
from .split_data import split_nist_dippr
from .remove_salts import remove_salts
from .remove_not_liquid import remove_not_liquid
from .T_logV_correlation import drop_flagged_data
from .remove_inconsistent import remove_inconsistent
from .pchip_interpolation import pchip_interpolation
from .write_data import write_data


warnings.filterwarnings("ignore")


def prepare_dataset(input_args):
    """
    Wrapper for data curation pipeline.
    """

    start = time.time()

    add_paths_to_args(input_args)
    
    print("Loading Data: ---")
    nist_knovel_all = load_data(input_args)
    print("Time Elapsed: {:.2f}".format(time.time() - start))

    print("Splitting Data: ---")
    test_mols = split_nist_dippr(nist_knovel_all, input_args)
    print("Total Time Elapsed: {:.2f}".format(time.time() - start))

    print("Removing Salts: ---")
    nist_knovel_all = remove_salts(nist_knovel_all, test_mols)
    print("Total Time Elapsed: {:.2f}".format(time.time() - start))

    print("Removing Non-Liquid Compounds: ---")
    nist_knovel_all = remove_not_liquid(nist_knovel_all,test_mols,input_args)
    print("Time Elapsed: {}".format(time.time()-start))

    print("Removing Compounds Based on Viscosity/Temperature Correlation: ---")
    nist_knovel_all = drop_flagged_data(nist_knovel_all, test_mols)
    print("Total Time Elapsed: {:.2f}".format(time.time() - start))

    print("Removing Inconsistent Data: ---")
    nist_knovel_all = remove_inconsistent(nist_knovel_all, test_mols, input_args)
    print("Total Time Elapsed: {:.2f}".format(time.time() - start))

    print("Combining Data using PCHIP Interpolation: ---")
    nist_knovel_all = pchip_interpolation(nist_knovel_all, test_mols)
    print("Total Time Elapsed: {}".format(time.time() - start))

    print("Writing Dataset to {}: ---".format(input_args["out_path"]))
    write_data(nist_knovel_all, test_mols, input_args)

    print("Total Time Elapsed: {:.2f}".format(time.time() - start))

def load_model(args):
    if 'checkpoint_dir' in args.keys():
        return mixprop_model(args['checkpoint_dir'])
    else:
        path = str(Path(__file__).absolute())
        path = '/'.join(path.split('/')[:-1])
        checkpoint_dir = os.path.join(path,'pretrained_models/nist_dippr_model/nist_dippr_model')
        
        print('Loading models from {}'.format(checkpoint_dir))
        
        return mixprop_model(checkpoint_dir)

curation_args = {
    "NIST": "pretrained_models/nist_dippr_source/NIST_Visc_Data.csv",  # Path to NIST data
    "DIPPR": "pretrained_models/nist_dippr_source/logV_bp_mp.csv",  # Path to DIPPR data
    "bp_pred_path": "pretrained_models/bp_data/bp_pred.csv",  # Path to boiling point data (already predicted)
    "mp_pred_path": "pretrained_models/mp_data/mp_pred.csv",  # Path to melting point data (already predicted)
    "dummy_mol": "CCO",  # Dummy compound for mixing in DIPPR pure component data
    "dippr_ref_T": 298,  # DIPPR temperature
    "test_split": 0.2,  # Fraction to hold out for testing
    "thresh_pure": 0.025,  # Settings for inconsistent pure data screening
    "thresh_logV": 0.5,  # Settings for inconsistent pure data screening
    "out_path":"." # Location for files to be written to
}
def add_paths_to_args(args):
    path = str(Path(__file__).absolute())
    path = '/'.join(path.split('/')[:-2])
    source_path = os.path.join(path,'pretrained_models')

    if 'NIST' not in args.keys():
        nist_path = os.path.join(source_path,'nist_dippr_source/NIST_Visc_Data.csv')
        args['NIST'] = nist_path
        print(print('Loading data from {}'.format(nist_path)))

    if 'DIPPR' not in args.keys():
        dippr_path = os.path.join(source_path,'nist_dippr_source/logV_bp_mp.csv')
        args['DIPPR'] = dippr_path
        print(print('Loading data from {}'.format(dippr_path)))
 
    if 'bp_pred_path' not in args.keys():
        bp_path = os.path.join(source_path,'bp_data/bp_pred.csv')
        args['bp_pred_path'] = bp_path
        print(print('Loading data from {}'.format(bp_path)))

    if 'mp_pred_path' not in args.keys():
        mp_path = os.path.join(source_path,'mp_data/mp_pred.csv')
        args['mp_pred_path'] = mp_path
        print(print('Loading data from {}'.format(mp_path)))
        
    return args
    



import csv
import os
import pickle
from typing import List

import numpy as np
import pandas as pd
from rdkit.Chem import PandasTools


def save_features(path: str, features: List[np.ndarray]) -> None:
    """
    Saves features to a compressed :code:`.npz` file with array name "features".

    :param path: Path to a :code:`.npz` file where the features will be saved.
    :param features: A list of 1D numpy arrays containing the features for molecules.
    """
    np.savez_compressed(path, features=features)


def load_features(path: str) -> np.ndarray:
    """
    Loads features saved in a variety of formats.

    Supported formats:

    * :code:`.npz` compressed (assumes features are saved with name "features")
    * .npy
    * :code:`.csv` / :code:`.txt` (assumes comma-separated features with a header and with one line per molecule)
    * :code:`.pkl` / :code:`.pckl` / :code:`.pickle` containing a sparse numpy array

    .. note::

       All formats assume that the SMILES loaded elsewhere in the code are in the same
       order as the features loaded here.

    :param path: Path to a file containing features.
    :return: A 2D numpy array of size :code:`(num_molecules, features_size)` containing the features.
    """
    extension = os.path.splitext(path)[1]

    if extension == '.npz':
        features = np.load(path)['features']
    elif extension == '.npy':
        features = np.load(path)
    elif extension in ['.csv', '.txt']:
        with open(path) as f:
            reader = csv.reader(f)
            next(reader)  # skip header
            features = np.array([[float(value) for value in row] for row in reader])
    elif extension in ['.pkl', '.pckl', '.pickle']:
        with open(path, 'rb') as f:
            features = np.array([np.squeeze(np.array(feat.todense())) for feat in pickle.load(f)])
    else:
        raise ValueError(f'Features path extension {extension} not supported.')

    return features


def load_valid_atom_or_bond_features(path: str, smiles: List[str]) -> List[np.ndarray]:
    """
    Loads features saved in a variety of formats.

    Supported formats:

    * :code:`.npz` descriptors are saved as 2D array for each molecule in the order of that in the data.csv
    * :code:`.pkl` / :code:`.pckl` / :code:`.pickle` containing a pandas dataframe with smiles as index and numpy array of descriptors as columns
    * :code:'.sdf' containing all mol blocks with descriptors as entries

    :param path: Path to file containing atomwise features.
    :return: A list of 2D array.
    """

    extension = os.path.splitext(path)[1]

    if extension == '.npz':
        container = np.load(path)
        features = [container[key] for key in container]

    elif extension in ['.pkl', '.pckl', '.pickle']:
        features_df = pd.read_pickle(path)
        if features_df.iloc[0, 0].ndim == 1:
            features = features_df.apply(lambda x: np.stack(x.tolist(), axis=1), axis=1).tolist()
        elif features_df.iloc[0, 0].ndim == 2:
            features = features_df.apply(lambda x: np.concatenate(x.tolist(), axis=1), axis=1).tolist()
        else:
            raise ValueError(f'Atom/bond descriptors input {path} format not supported')

    elif extension == '.sdf':
        features_df = PandasTools.LoadSDF(path).drop(['ID', 'ROMol'], axis=1).set_index('SMILES')

        features_df = features_df[~features_df.index.duplicated()]

        # locate atomic descriptors columns
        features_df = features_df.iloc[:, features_df.iloc[0, :].apply(lambda x: isinstance(x, str) and ',' in x).to_list()]
        features_df = features_df.reindex(smiles)
        if features_df.isnull().any().any():
            raise ValueError('Invalid custom atomic descriptors file, Nan found in data')

        features_df = features_df.applymap(lambda x: np.array(x.replace('\r', '').replace('\n', '').split(',')).astype(float))

        features = features_df.apply(lambda x: np.stack(x.tolist(), axis=1), axis=1).tolist()

    else:
        raise ValueError(f'Extension "{extension}" is not supported.')

    return features



from .utils import report_stats


def remove_salts(nist_knovel_all, test_mols):
    """
    Remove any datapoint containing a SMILES string with multiple molecules present.
    """

    nist_knovel_all = nist_knovel_all[
        nist_knovel_all["MOL_1"].apply(lambda x: "." not in x)
    ]
    nist_knovel_all = nist_knovel_all[
        nist_knovel_all["MOL_2"].apply(lambda x: "." not in x)
    ]

    report_stats(nist_knovel_all, test_mols)

    return nist_knovel_all



import numpy as np
import pandas as pd
from .utils import report_stats


def split_nist_dippr(nist_knovel_all, args):
    """
    Split the combined NIST/DIPPR dataset sequentially, first splitting the non-pure data and second splitting
    the pure data.
    """

    ## Split non-pure data:
    nist_knovel_all["pure"] = (nist_knovel_all["MolFrac_1"] == 0.0) | (
        nist_knovel_all["MolFrac_1"] == 1.0
    )
    test_split = args["test_split"]
    subset = nist_knovel_all[~nist_knovel_all["pure"]]
    test_count = int(test_split * len(subset))

    subset_mols = pd.DataFrame(
        list(set(list(subset["MOL_1"].values) + list(subset["MOL_2"].values)))
    )

    test_size = 0
    i = 175
    while test_size < test_count:
        # print(test_size,test_count)
        test_mols = subset_mols.sample(n=i)[0].values
        if ("O" not in test_mols) & (
            "CCO" not in test_mols
        ):  # &('CCCCO' not in test_mols)&('CCCO' not in test_mols)&('c1ccccc1' not in test_mols)&('CO' not in test_mols):
            test_size = np.sum(
                nist_knovel_all["MOL_1"].apply(lambda smi: smi in test_mols)
                | nist_knovel_all["MOL_2"].apply(lambda smi: smi in test_mols)
            )
        i += 1

        nonpure_test_mols = test_mols

    ## Split pure data:
    subset = nist_knovel_all[nist_knovel_all["pure"]]
    test_count = int(test_split * len(subset))

    subset_mols = pd.DataFrame(
        list(set(list(subset["MOL_1"].values) + list(subset["MOL_2"].values)))
    )

    test_size = 0
    i = 1
    while test_size < test_count:
        test_mols = list(subset_mols.sample(n=i)[0].values)

        test_mols = list(set(test_mols + list(nonpure_test_mols)))

        if ("O" not in test_mols) & (
            "CCO" not in test_mols
        ):  # &('CCCCO' not in test_mols)&('CCCO' not in test_mols)&('c1ccccc1' not in test_mols)&('CO' not in test_mols):
            test_size = np.sum(
                nist_knovel_all["MOL_1"].apply(lambda smi: smi in test_mols)
                | nist_knovel_all["MOL_2"].apply(lambda smi: smi in test_mols)
            )
        i += 1

    report_stats(nist_knovel_all, test_mols)

    return test_mols



import logging
from typing import Callable

from tensorboardX import SummaryWriter
import torch
import torch.nn as nn
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from tqdm import tqdm

from mixprop.args import TrainArgs
from mixprop.data import MoleculeDataLoader, MoleculeDataset
from mixprop.models import MoleculeModel
from mixprop.nn_utils import compute_gnorm, compute_pnorm, NoamLR


def train(model: MoleculeModel,
          data_loader: MoleculeDataLoader,
          loss_func: Callable,
          optimizer: Optimizer,
          scheduler: _LRScheduler,
          args: TrainArgs,
          n_iter: int = 0,
          logger: logging.Logger = None,
          writer: SummaryWriter = None) -> int:
    """
    Trains a model for an epoch.

    :param model: A :class:`~mixprop.models.model.MoleculeModel`.
    :param data_loader: A :class:`~mixprop.data.data.MoleculeDataLoader`.
    :param loss_func: Loss function.
    :param optimizer: An optimizer.
    :param scheduler: A learning rate scheduler.
    :param args: A :class:`~mixprop.args.TrainArgs` object containing arguments for training the model.
    :param n_iter: The number of iterations (training examples) trained on so far.
    :param logger: A logger for recording output.
    :param writer: A tensorboardX SummaryWriter.
    :return: The total number of iterations (training examples) trained on so far.
    """
    debug = logger.debug if logger is not None else print

    model.train()
    loss_sum = iter_count = 0

    for batch in tqdm(data_loader, total=len(data_loader), leave=False):
        # Prepare batch
        batch: MoleculeDataset
        mol_batch, features_batch, target_batch, atom_descriptors_batch, atom_features_batch, bond_features_batch, data_weights_batch = \
            batch.batch_graph(), batch.features(), batch.targets(), batch.atom_descriptors(), \
            batch.atom_features(), batch.bond_features(), batch.data_weights()

        mask = torch.tensor([[x is not None for x in tb] for tb in target_batch], dtype=torch.bool) # shape(batch, tasks)
        targets = torch.tensor([[0 if x is None else x for x in tb] for tb in target_batch]) # shape(batch, tasks)

        if args.target_weights is not None:
            target_weights = torch.tensor(args.target_weights).unsqueeze(0) # shape(1,tasks)
        else:
            target_weights = torch.ones(targets.shape[1]).unsqueeze(0)
        data_weights = torch.tensor(data_weights_batch).unsqueeze(1) # shape(batch,1)

        if args.loss_function == 'bounded_mse':
            lt_target_batch = batch.lt_targets() # shape(batch, tasks)
            gt_target_batch = batch.gt_targets() # shape(batch, tasks)
            lt_target_batch = torch.tensor(lt_target_batch)
            gt_target_batch = torch.tensor(gt_target_batch)

        # Run model
        model.zero_grad()
        preds = model(mol_batch, features_batch, atom_descriptors_batch, atom_features_batch, bond_features_batch)

        # Move tensors to correct device
        torch_device = preds.device
        mask = mask.to(torch_device)
        targets = targets.to(torch_device)
        target_weights = target_weights.to(torch_device)
        data_weights = data_weights.to(torch_device)
        if args.loss_function == 'bounded_mse':
            lt_target_batch = lt_target_batch.to(torch_device)
            gt_target_batch = gt_target_batch.to(torch_device)

        # Calculate losses
        if args.loss_function == 'mcc' and args.dataset_type == 'classification':
            loss = loss_func(preds, targets, data_weights, mask) *target_weights.squeeze(0)
        elif args.loss_function == 'mcc': # multiclass dataset type
            targets = targets.long()
            target_losses = []
            for target_index in range(preds.size(1)):
                target_loss = loss_func(preds[:, target_index, :], targets[:, target_index], data_weights, mask[:, target_index]).unsqueeze(0)
                target_losses.append(target_loss)
            loss = torch.cat(target_losses).to(torch_device) * target_weights.squeeze(0)
        elif args.dataset_type == 'multiclass':
            targets = targets.long()
            target_losses = []
            for target_index in range(preds.size(1)):
                target_loss = loss_func(preds[:, target_index, :], targets[:, target_index]).unsqueeze(1)
                target_losses.append(target_loss)
            loss = torch.cat(target_losses, dim=1).to(torch_device) * target_weights * data_weights * mask
        elif args.dataset_type == 'spectra':
            loss = loss_func(preds, targets, mask) * target_weights * data_weights * mask
        elif args.loss_function == 'bounded_mse':
            loss = loss_func(preds, targets, lt_target_batch, gt_target_batch) * target_weights * data_weights * mask
        else:
            loss = loss_func(preds, targets) * target_weights * data_weights * mask
        loss = loss.sum() / mask.sum()

        loss_sum += loss.item()
        iter_count += 1

        loss.backward()
        if args.grad_clip:
            nn.utils.clip_grad_norm_(model.parameters(), args.grad_clip)
        optimizer.step()

        if isinstance(scheduler, NoamLR):
            scheduler.step()

        n_iter += len(batch)

        # Log and/or add to tensorboard
        if (n_iter // args.batch_size) % args.log_frequency == 0:
            lrs = scheduler.get_lr()
            pnorm = compute_pnorm(model)
            gnorm = compute_gnorm(model)
            loss_avg = loss_sum / iter_count
            loss_sum = iter_count = 0

            lrs_str = ', '.join(f'lr_{i} = {lr:.4e}' for i, lr in enumerate(lrs))
            debug(f'Loss = {loss_avg:.4e}, PNorm = {pnorm:.4f}, GNorm = {gnorm:.4f}, {lrs_str}')

            if writer is not None:
                writer.add_scalar('train_loss', loss_avg, n_iter)
                writer.add_scalar('param_norm', pnorm, n_iter)
                writer.add_scalar('gradient_norm', gnorm, n_iter)
                for i, lr in enumerate(lrs):
                    writer.add_scalar(f'learning_rate_{i}', lr, n_iter)

    return n_iter



from mixprop.args import HyperoptArgs
import os
import pickle
from typing import List, Dict
import csv
import json

from hyperopt import Trials

from mixprop.constants import HYPEROPT_SEED_FILE_NAME
from mixprop.utils import makedirs

def merge_trials(trials: Trials, new_trials_data: List[Dict]) -> Trials:
    """
    Merge a hyperopt trials object with the contents of another hyperopt trials object.

    :param trials: A hyperopt trials object containing trials data, organized into hierarchical dictionaries.
    :param trials_data: The contents of a hyperopt trials object, `Trials.trials`.
    :return: A hyperopt trials object, merged from the two inputs.
    """
    max_tid = 0
    if len(trials.trials) > 0:
        max_tid = max([trial['tid'] for trial in trials.trials])

    for trial in new_trials_data:
        tid = trial['tid'] + max_tid + 1 #trial id needs to be unique among this list of ids.
        hyperopt_trial = Trials().new_trial_docs(
                tids=[None],
                specs=[None],
                results=[None],
                miscs=[None])
        hyperopt_trial[0] = trial
        hyperopt_trial[0]['tid'] = tid
        hyperopt_trial[0]['misc']['tid'] = tid
        for key in hyperopt_trial[0]['misc']['idxs'].keys():
            hyperopt_trial[0]['misc']['idxs'][key] = [tid]
        trials.insert_trial_docs(hyperopt_trial) 
        trials.refresh()
    return trials


def load_trials(dir_path: str, previous_trials: Trials = None) -> Trials:
    """
    Load in trials from each pickle file in the hyperopt checkpoint directory.
    Checkpoints are newly loaded in at each iteration to allow for parallel entries
    into the checkpoint folder by independent hyperoptimization instances.

    :param dir_path: Path to the directory containing hyperopt checkpoint files.
    :param previous_trials: Any previously generated trials objects that the loaded trials will be merged with.
    :return: A trials object containing the merged trials from all checkpoint files.
    """

    # List out all the pickle files in the hyperopt checkpoint directory
    hyperopt_checkpoint_files = [os.path.join(dir_path, path) for path in os.listdir(dir_path) if '.pkl' in path]

    # Load hyperopt trials object from each file
    loaded_trials = Trials()
    if previous_trials is not None:
        loaded_trials = merge_trials(loaded_trials, previous_trials.trials)

    for path in hyperopt_checkpoint_files:
        with open(path,'rb') as f:
            trial = pickle.load(f)
            loaded_trials = merge_trials(loaded_trials, trial.trials)
    
    return loaded_trials


def save_trials(dir_path: str, trials: Trials, hyperopt_seed: int) -> None:
    """
    Saves hyperopt trial data as a `.pkl` file.

    :param dir_path: Path to the directory containing hyperopt checkpoint files.
    :param trials: A trials object containing information on a completed hyperopt iteration.
    """
    new_fname = f'{hyperopt_seed}.pkl'
    existing_files = os.listdir(dir_path)
    if new_fname in existing_files:
        raise ValueError(f'When saving trial with unique seed {hyperopt_seed}, found that a trial with this seed already exists.')
    pickle.dump(trials, open(os.path.join(dir_path, new_fname), 'wb'))


def get_hyperopt_seed(seed: int, dir_path: str) -> int:
    """
    Assigns a seed for hyperopt calculations. Each iteration will start with a different seed.

    :param seed: The initial attempted hyperopt seed.
    :param dir_path: Path to the directory containing hyperopt checkpoint files.
    :return: An integer for use as hyperopt random seed.
    """

    seed_path = os.path.join(dir_path,HYPEROPT_SEED_FILE_NAME)
    
    seeds = []
    if os.path.exists(seed_path):
        with open(seed_path, 'r') as f:
            seed_line = next(f)
            seeds.extend(seed_line.split())
    else:
        makedirs(seed_path, isfile=True)

    seeds = [int(sd) for sd in seeds]

    while seed in seeds:
        seed += 1
    seeds.append(seed)

    write_line = " ".join(map(str, seeds)) + '\n'

    with open(seed_path, 'w') as f:
        f.write(write_line)
    
    return seed


def load_manual_trials(manual_trials_dirs: List[str], param_keys: List[str], hyperopt_args: HyperoptArgs) -> Trials:
    """
    Function for loading in manual training runs as trials for inclusion in hyperparameter search.
    Trials must be consistent in all arguments with trials that would be generated in hyperparameter optimization.

    :param manual_trials_dirs: A list of paths to save directories for the manual trials, as would include test_scores.csv and args.json.
    :param param_keys: A list of the parameters included in the hyperparameter optimization.
    :param hyperopt_args: The arguments for the hyperparameter optimization job.
    :return: A hyperopt trials object including all the loaded manual trials.
    """
    matching_args = [ # manual trials must occupy the same space as the hyperparameter optimization search. This is a non-extensive list of arguments to check to see if they are consistent.
        'number_of_molecules',
        'aggregation',
        'num_folds',
        'ensemble_size',
        'max_lr',
        'init_lr',
        'final_lr',
        'activation',
        'metric',
        'bias',
        'epochs',
        'explicit_h',
        'reaction',
        'split_type',
        'warmup_epochs',
    ]

    manual_trials_data = []
    for i, trial_dir in enumerate(manual_trials_dirs):

        # Extract trial data from test_scores.csv
        with open(os.path.join(trial_dir, 'test_scores.csv')) as f:
            reader=csv.reader(f)
            next(reader)
            read_line=next(reader)
        mean_score = float(read_line[1])
        std_score = float(read_line[2])
        loss = (1 if hyperopt_args.minimize_score else -1) * mean_score

        # Extract argument data from args.json
        with open(os.path.join(trial_dir, 'args.json')) as f:
            trial_args = json.load(f)

        # Check for differences in manual trials and hyperopt space
        if 'hidden_size' in param_keys:
            if trial_args['hidden_size'] != trial_args['ffn_hidden_size']:
                raise ValueError(f'The manual trial in {trial_dir} has a hidden_size {trial_args["hidden_size"]} '
                f'that does not match its ffn_hidden_size {trial_args["ffn_hidden_size"]}, as it would in hyperparameter search.')
        for arg in matching_args:
            if arg not in param_keys:
                if getattr(hyperopt_args,arg) != trial_args[arg]:
                    raise ValueError(f'Manual trial {trial_dir} has different training argument {arg} than the hyperparameter optimization search trials.')

        # Construct data dict
        param_dict = {key: trial_args[key] for key in param_keys}
        vals_dict = {key: [param_dict[key]] for key in param_keys}
        idxs_dict = {key: [i] for key in param_keys}
        results_dict = {
            'loss': loss,
            'status': 'ok',
            'mean_score': mean_score,
            'std_score': std_score,
            'hyperparams': param_dict,
            'num_params': 0,
        }
        misc_dict = {
            'tid': i,
            'cmd': ('domain_attachment', 'FMinIter_Domain'),
            'workdir': None,
            'idxs': idxs_dict,
            'vals': vals_dict,
        }
        trial_data = {
            'state': 2,
            'tid': i,
            'spec': None,
            'result': results_dict,
            'misc': misc_dict,
            'exp_key': None,
            'owner': None,
            'version': 0,
            'book_time': None,
            'refresh_time': None,
        }
        manual_trials_data.append(trial_data)

    trials = Trials()
    trials = merge_trials(trials=trials, new_trials_data=manual_trials_data)
    return trials



import numpy as np
import pandas as pd
from scipy.stats import pearsonr

from .utils import report_stats


def flag_data(subset):
    """
    Create a list of indices corresponding to possibly erroneous data, based on
    spurious viscosity-temperature correlations.
    """
    index_drop_list = []

    pr = pearsonr(subset["T"], subset["logV"])[0]
    std = np.std(subset["logV"])

    def criteria(pr, std):
        return (pr > -0.75) & (std > 0.15)

    i = 0
    iters = 0
    max_iters = 50
    while (pr > -0.75) & (std > 0.15):
        if i >= len(subset.index):
            i = 0
            iters += 1
        if iters > max_iters:
            break
        index = subset.index[i]
        subset_dropped = subset.drop(index=index)
        pr_dropped = pearsonr(subset_dropped["T"], subset_dropped["logV"])[0]
        std_dropped = np.std(subset_dropped["logV"])

        # If dropping the data brings the series within our criteria, then drop it:
        if not criteria(pr_dropped, std_dropped):
            subset = subset_dropped
            index_drop_list.append(index)

        if pr - pr_dropped > 0.1:
            subset = subset_dropped
            index_drop_list.append(index)

        # Update pr and std and i
        pr = pearsonr(subset["T"], subset["logV"])[0]
        std = np.std(subset["logV"])
        i += 1

    index_drop_list = list(set(index_drop_list))

    return index_drop_list


def drop_flagged_data(data, test_mols):
    """
    Remove potentially spurious data based on unusual viscosity-temperature correlations.
    """

    pure1 = data[data["MolFrac_1"] == 1]
    pure1["TargetMol"] = pure1["MOL_1"]
    pure2 = data[data["MolFrac_1"] == 0]
    pure2["TargetMol"] = pure2["MOL_2"]

    pure_all = pd.concat((pure1, pure2))
    unique_mols = pure_all["TargetMol"].drop_duplicates().values

    correlations = []
    std = []
    for smi in unique_mols:
        subset = pure_all[pure_all["TargetMol"] == smi]
        if len(subset["T"].drop_duplicates()) > 1:
            correlations.append(pearsonr(subset["T"], subset["logV"])[0])
            std.append(np.std(subset["logV"]))
        else:
            correlations.append(None)
            std.append(None)

    pure_Tdep = pd.DataFrame(
        {"SMILES": unique_mols, "PR": correlations, "std": std}
    ).dropna()

    suspect_list = pure_Tdep[(pure_Tdep["PR"] > -0.75) & (pure_Tdep["std"] > 0.15)][
        "SMILES"
    ].values

    index_drop_list = []
    for smi in suspect_list:
        subset = pure_all[pure_all["TargetMol"] == smi]
        index_drop_list.extend(flag_data(subset))

    for index in index_drop_list:
        smi1, smi2, T = pure_all.loc[index][["MOL_1", "MOL_2", "T"]]

        index2drop = data[
            (data["MOL_1"] == smi1) & (data["MOL_2"] == smi2) & (data["T"] == T)
        ].index
        try:
            data = data.drop(index=index2drop)
        except:
            print("Data has already been dropped")

    report_stats(data, test_mols)

    return data



from typing import Callable

import torch
import torch.nn as nn

from mixprop.args import TrainArgs


def get_loss_func(args: TrainArgs) -> Callable:
    """
    Gets the loss function corresponding to a given dataset type.

    :param args: Arguments containing the dataset type ("classification", "regression", or "multiclass").
    :return: A PyTorch loss function.
    """

    # Nested dictionary of the form {dataset_type: {loss_function: loss_function callable}}
    supported_loss_functions ={
        'regression':{
            'mse': nn.MSELoss(reduction='none'),
            'bounded_mse': bounded_mse_loss
        },
        'classification':{
            'binary_cross_entropy': nn.BCEWithLogitsLoss(reduction='none'),
            'mcc': mcc_class_loss,
        },
        'multiclass':{
            'cross_entropy': nn.CrossEntropyLoss(reduction='none'),
            'mcc': mcc_multiclass_loss,
        },
        'spectra':{
            'sid': sid_loss,
            'wasserstein': wasserstein_loss,
        }
    }

    # Error if no loss function supported
    if args.dataset_type not in supported_loss_functions.keys():
        raise ValueError(f'Dataset type "{args.dataset_type}" not supported.')

    # Return loss function if it is represented in the supported_loss_functions dictionary
    loss_function = supported_loss_functions.get(args.dataset_type, dict()).get(args.loss_function, None)

    if loss_function is not None:
        return loss_function

    else:
        raise ValueError(f'Loss function "{args.loss_function}" not supported with dataset type {args.dataset_type}. \
            Available options for that dataset type are {supported_loss_functions[args.dataset_type].keys()}.')


def bounded_mse_loss(predictions: torch.tensor, targets: torch.tensor, less_than_target: torch.tensor, greater_than_target: torch.tensor) -> torch.tensor:
    """
    Loss function for use with regression when some targets are presented as inequalities.

    :param predictions: Model predictions with shape(batch_size, tasks).
    :param targets: Target values with shape(batch_size, tasks).
    :param less_than_target: A tensor with boolean values indicating whether the target is a less-than inequality.
    :param greater_than_target: A tensor with boolean values indicating whether the target is a greater-than inequality.
    :return: A tensor containing loss values of shape(batch_size, tasks).
    """
    predictions = torch.where(
        torch.logical_and(predictions < targets, less_than_target),
        targets,
        predictions
    )

    predictions = torch.where(
        torch.logical_and(predictions > targets, greater_than_target),
        targets,
        predictions
    )
    
    return nn.functional.mse_loss(predictions, targets, reduction='none')


def mcc_class_loss(predictions: torch.tensor, targets: torch.tensor, data_weights: torch.tensor, mask: torch.tensor) -> torch.tensor:
    """
    A classification loss using a soft version of the Matthews Correlation Coefficient.

    :param predictions: Model predictions with shape(batch_size, tasks).
    :param targets: Target values with shape(batch_size, tasks).
    :param data_weights: A tensor with float values indicating how heavily to weight each datapoint in training with shape(batch_size, 1)
    :param mask: A tensor with boolean values indicating whether the loss for this prediction is considered in the gradient descent with shape(batch_size, tasks).
    :return: A tensor containing loss values of shape(tasks).
    """
    # shape(batch, tasks)
    # (TP*TN-FP*FN)/sqrt((TP+FP)*(TP+FN)*(TN+FP)*(TN+FN))
    torch_device = predictions.device
    TP = torch.sum(targets * predictions * data_weights * mask, axis = 0).to(torch_device)
    FP = torch.sum((1 - targets) * predictions * data_weights * mask, axis = 0).to(torch_device)
    FN = torch.sum(targets * (1 - predictions) * data_weights * mask, axis = 0).to(torch_device)
    TN = torch.sum((1 - targets) * (1 - predictions) * data_weights * mask, axis = 0).to(torch_device)
    loss = 1 - ((TP*TN-FP*FN)/torch.sqrt((TP+FP)*(TP+FN)*(TN+FP)*(TN+FN)))
    loss = loss.to(torch_device)
    return loss


def mcc_multiclass_loss(predictions: torch.tensor, targets: torch.tensor, data_weights: torch.tensor, mask: torch.tensor) -> torch.tensor:
    """
    A multiclass loss using a soft version of the Matthews Correlation Coefficient. Multiclass definition follows the version in sklearn documentation.

    :param predictions: Model predictions with shape(batch_size, classes).
    :param targets: Target values with shape(batch_size).
    :param data_weights: A tensor with float values indicating how heavily to weight each datapoint in training with shape(batch_size, 1)
    :param mask: A tensor with boolean values indicating whether the loss for this prediction is considered in the gradient descent with shape(batch_size).
    :return: A tensor value for the loss.
    """
    # targets shape (batch)
    # preds shape(batch, classes)
    torch_device = predictions.device
    mask = mask.unsqueeze(1)
    bin_targets = torch.zeros_like(predictions, device=torch_device)
    bin_targets[torch.arange(predictions.shape[0]), targets] = 1
    c = torch.sum(predictions * bin_targets * data_weights * mask).to(torch_device)
    s = torch.sum(predictions * data_weights * mask).to(torch_device)
    pt = torch.sum(torch.sum(predictions * data_weights * mask, axis=0) * torch.sum(bin_targets * data_weights * mask, axis=0)).to(torch_device)
    p2 = torch.sum(torch.sum(predictions * data_weights * mask, axis=0)**2).to(torch_device)
    t2 = torch.sum(torch.sum(bin_targets * data_weights * mask, axis=0)**2).to(torch_device)
    loss = 1 - (c * s - pt) / torch.sqrt((s**2 - p2)*(s**2 - t2))
    loss = loss.to(torch_device)
    return loss


def sid_loss(model_spectra: torch.tensor, target_spectra: torch.tensor, mask: torch.tensor, threshold: float = None) -> torch.tensor:
    """
    Loss function for use with spectra data type.

    :param model_spectra: The predicted spectra output from a model with shape (batch_size,spectrum_length).
    :param target_spectra: The target spectra with shape (batch_size,spectrum_length). Values must be normalized so that each spectrum sums to 1.
    :param mask: Tensor with boolean indications of where the spectrum output should not be excluded with shape (batch_size,spectrum_length).
    :param threshold: Loss function requires that values are positive and nonzero. Values below the threshold will be replaced with the threshold value.
    :return: A tensor containing loss values for the batch with shape (batch_size,spectrum_length).
    """
    # Move new tensors to torch device
    torch_device = model_spectra.device

    # Normalize the model spectra before comparison
    zero_sub = torch.zeros_like(model_spectra, device=torch_device)
    one_sub = torch.ones_like(model_spectra, device=torch_device)
    if threshold is not None:
        threshold_sub = torch.full(model_spectra.shape, threshold, device=torch_device)
        model_spectra = torch.where(model_spectra < threshold, threshold_sub, model_spectra)
    model_spectra = torch.where(mask, model_spectra, zero_sub)
    sum_model_spectra = torch.sum(model_spectra, axis=1, keepdim=True)
    model_spectra = torch.div(model_spectra, sum_model_spectra)

    # Calculate loss value
    target_spectra = torch.where(mask, target_spectra, one_sub)
    model_spectra = torch.where(mask, model_spectra, one_sub) # losses in excluded regions will be zero because log(1/1) = 0. 
    loss = torch.mul(torch.log(torch.div(model_spectra, target_spectra)), model_spectra) \
        + torch.mul(torch.log(torch.div(target_spectra, model_spectra)), target_spectra)
    loss = loss.to(torch_device)

    return loss


def wasserstein_loss(model_spectra: torch.tensor, target_spectra: torch.tensor, mask: torch.tensor, threshold: float = None) -> torch.tensor:
    """
    Loss function for use with spectra data type. This loss assumes that values are evenly spaced.

    :param model_spectra: The predicted spectra output from a model with shape (batch_size,spectrum_length).
    :param target_spectra: The target spectra with shape (batch_size,spectrum_length). Values must be normalized so that each spectrum sums to 1.
    :param mask: Tensor with boolian indications of where the spectrum output should not be excluded with shape (batch_size,spectrum_length).
    :param threshold: Loss function requires that values are positive and nonzero. Values below the threshold will be replaced with the threshold value.
    :return: A tensor containing loss values for the batch with shape (batch_size,spectrum_length).
    """
    # Move new tensors to torch device
    torch_device = model_spectra.device

    # Normalize the model spectra before comparison
    zero_sub = torch.zeros_like(model_spectra, device=torch_device)
    if threshold is not None:
        threshold_sub = torch.full(model_spectra.shape,threshold, device=torch_device)
        model_spectra = torch.where(model_spectra < threshold, threshold_sub, model_spectra)
    model_spectra = torch.where(mask, model_spectra, zero_sub)
    sum_model_spectra = torch.sum(model_spectra, axis=1, keepdim=True)
    model_spectra = torch.div(model_spectra, sum_model_spectra)

    # Calculate loss value
    target_cum = torch.cumsum(target_spectra,axis=1).to(torch_device)
    model_cum = torch.cumsum(model_spectra,axis=1).to(torch_device)
    loss = torch.abs(target_cum - model_cum)
    loss = loss.to(torch_device)

    return loss



import numpy as np
import pandas as pd
from scipy.interpolate import PchipInterpolator

from .utils import series_std, report_stats


def pchip_interpolation(nist_knovel_all, test_mols):
    """
    Interpolate between data for a given mixture/temperature combination and average over data.
    """

    pairs = []
    one_off_data = []
    for smi1, smi2, id1, id2, T in (
        nist_knovel_all[["MOL_1", "MOL_2", "ID_1", "ID_2", "T"]]
        .drop_duplicates()
        .values
    ):
        subset = nist_knovel_all[
            (nist_knovel_all["MOL_1"] == smi1)
            & (nist_knovel_all["MOL_2"] == smi2)
            & (nist_knovel_all["T"] == T)
        ]

        pair_entry = {}
        pair_entry["MOL_1"], pair_entry["MOL_2"] = smi1, smi2
        pair_entry["ID_1"], pair_entry["ID_2"] = id1, id2

        frac_min = min(subset["MolFrac_1"])
        frac_max = max(subset["MolFrac_1"])
        xrange = np.arange(frac_min, frac_max + 0.1, 0.1)
        ydict = {round(x, 2): [] for x in xrange}

        pair_entry["ref_ids"] = subset.Ref_ID.drop_duplicates().values
        pair_entry["T"] = T

        # For each experimental setup:
        for smi1, smi2, ref_id in (
            subset[["MOL_1", "MOL_2", "Ref_ID"]].drop_duplicates().values
        ):
            sub_subset = subset[
                (subset["MOL_1"] == smi1)
                & (subset["MOL_2"] == smi2)
                & (subset["Ref_ID"] == ref_id)
            ].sort_values("MolFrac_1")
            sub_subset.groupby("MolFrac_1").mean()["logV"]

            # We will deal with them by averaging:
            fractions = sub_subset["MolFrac_1"].drop_duplicates()
            values = sub_subset.groupby("MolFrac_1").mean()["logV"]
            if len(fractions) > 1:

                pchip = PchipInterpolator(fractions, values)

                # Find min and max fractions for this experiment:
                sub_frac_min = min(sub_subset["MolFrac_1"])
                sub_frac_max = max(sub_subset["MolFrac_1"])

                for xval in ydict.keys():
                    if (xval >= sub_frac_min) & (xval <= sub_frac_max):
                        ydict[xval].append(pchip(xval).mean())
            else:
                one_off_data.append(
                    (
                        smi1,
                        smi2,
                        ref_id,
                        fractions.values,
                        values.values,
                        T,
                        sub_subset["Visc_Unc"].values[0],
                    )
                )

        pair_entry["Data"] = ydict
        pair_entry["Data_Avg"] = [
            list(ydict.keys()),
            [np.mean(ydict[x]) for x in ydict.keys()],
        ]
        pair_entry["Data_Std"] = [
            list(ydict.keys()),
            [np.std(ydict[x]) for x in ydict.keys()],
        ]
        pair_entry["Average Visc_Unc (mPas)"] = np.mean(sub_subset["Visc_Unc"] * 1000)
        pairs.append(pair_entry)

    pairs_df = pd.DataFrame(pairs)
    pairs_df["Series_Uncertainty"] = pairs_df["Data"].apply(series_std)
    pairs_df_clean = pairs_df

    pd_input = []
    mol_list = []
    for i in range(len(pairs_df_clean)):
        reported_unc = pairs_df_clean.iloc[i]["Average Visc_Unc (mPas)"]
        series_unc = pairs_df_clean.iloc[i]["Series_Uncertainty"]
        for frac, val in np.transpose(pairs_df_clean.iloc[i]["Data_Avg"]):
            pd_input_dict = {}
            pd_input_dict["MOL_1"], pd_input_dict["MOL_2"] = (
                pairs_df_clean.iloc[i]["MOL_1"],
                pairs_df_clean.iloc[i]["MOL_2"],
            )
            pd_input_dict["Ref_ID"] = pairs_df_clean.iloc[i]["ref_ids"]
            pd_input_dict["ID_1"], pd_input_dict["ID_2"] = (
                pairs_df_clean.iloc[i]["ID_1"],
                pairs_df_clean.iloc[i]["ID_2"],
            )
            pd_input_dict["MolFrac_1"] = frac
            pd_input_dict["logV"] = val
            pd_input_dict["T"] = pairs_df_clean.iloc[i]["T"]
            pd_input_dict["Avg_Reported_Unc"] = reported_unc
            pd_input_dict["Avg_Series_Unc"] = series_unc
            pd_input.append(pd_input_dict)
            mol_list.append(pd_input_dict["MOL_1"])
            mol_list.append(pd_input_dict["MOL_2"])
    ip_data = pd.DataFrame(pd_input)

    input_data = ip_data
    all_dicts = []
    for single_entry in one_off_data:
        single_dict = {
            "MOL_1": single_entry[0],
            "MOL_2": single_entry[1],
            "MolFrac_1": single_entry[3],
            "logV": single_entry[4],
            "T": single_entry[5],
            "Avg_Series_Unc": single_entry[6],
        }
        single_df = pd.DataFrame(single_dict)
        input_data = pd.concat((input_data, single_df))
        all_dicts.append(single_dict)

    input_data = input_data[
        [
            "MOL_1",
            "MOL_2",
            "MolFrac_1",
            "logV",
            "T",
            "Avg_Reported_Unc",
            "Avg_Series_Unc",
        ]
    ].drop_duplicates()

    report_stats(input_data, test_mols)

    return input_data



from collections import OrderedDict
import csv
from typing import List, Optional, Union, Tuple

import numpy as np
from tqdm import tqdm

from .predict import predict
from mixprop.spectra_utils import normalize_spectra, roundrobin_sid
from mixprop.args import PredictArgs, TrainArgs
from mixprop.data import get_data, get_data_from_smiles, MoleculeDataLoader, MoleculeDataset, StandardScaler
from mixprop.utils import load_args, load_checkpoint, load_scalers, makedirs, timeit, update_prediction_args
from mixprop.features import set_extra_atom_fdim, set_extra_bond_fdim, set_reaction, set_explicit_h, set_adding_hs, reset_featurization_parameters
from mixprop.models import MoleculeModel


def load_model(args: PredictArgs, generator: bool = False):
    """
    Function to load a model or ensemble of models from file. If generator is True, a generator of the respective model and scaler 
    objects is returned (memory efficient), else the full list (holding all models in memory, necessary for preloading).

    :param args: A :class:`~mixprop.args.PredictArgs` object containing arguments for
                 loading data and a model and making predictions.
    :param generator: A boolean to return a generator instead of a list of models and scalers.
    :return: A tuple of updated prediction arguments, training arguments, a list or generator object of models, a list or 
                 generator object of scalers, the number of tasks and their respective names.
    """
    print('Loading training args')
    train_args = load_args(args.checkpoint_paths[0])
    num_tasks, task_names = train_args.num_tasks, train_args.task_names

    update_prediction_args(predict_args=args, train_args=train_args)
    args: Union[PredictArgs, TrainArgs]

    # Load model and scalers
    models = (load_checkpoint(checkpoint_path, device=args.device) for checkpoint_path in args.checkpoint_paths)
    scalers = (load_scalers(checkpoint_path) for checkpoint_path in args.checkpoint_paths)
    if not generator:
        models = list(models)
        scalers = list(scalers)

    return args, train_args, models, scalers, num_tasks, task_names


def load_data(args: PredictArgs, smiles: List[List[str]]):
    """
    Function to load data from a list of smiles or a file.

    :param args: A :class:`~mixprop.args.PredictArgs` object containing arguments for
                 loading data and a model and making predictions.
    :param smiles: A list of list of smiles, or None if data is to be read from file
    :return: A tuple of a :class:`~mixprop.data.MoleculeDataset` containing all datapoints, a :class:`~mixprop.data.MoleculeDataset` containing only valid datapoints,
                 a :class:`~mixprop.data.MoleculeDataLoader` and a dictionary mapping full to valid indices. 
    """
    print('Loading data')
    if smiles is not None:
        full_data = get_data_from_smiles(
            smiles=smiles,
            skip_invalid_smiles=False,
            features_generator=args.features_generator
        )
    else:
        full_data = get_data(path=args.test_path, smiles_columns=args.smiles_columns, target_columns=[], ignore_columns=[],
                             skip_invalid_smiles=False, args=args, store_row=not args.drop_extra_columns)

    print('Validating SMILES')
    full_to_valid_indices = {}
    valid_index = 0
    for full_index in range(len(full_data)):
        if all(mol is not None for mol in full_data[full_index].mol):
            full_to_valid_indices[full_index] = valid_index
            valid_index += 1

    test_data = MoleculeDataset([full_data[i] for i in sorted(full_to_valid_indices.keys())])

    print(f'Test size = {len(test_data):,}')

    # Create data loader
    test_data_loader = MoleculeDataLoader(
        dataset=test_data,
        batch_size=args.batch_size,
        num_workers=args.num_workers
    )

    return full_data, test_data, test_data_loader, full_to_valid_indices


def set_features(args: PredictArgs, train_args: TrainArgs):
    """
    Function to set extra options.

    :param args: A :class:`~mixprop.args.PredictArgs` object containing arguments for
                 loading data and a model and making predictions.
    :param train_args: A :class:`~mixprop.args.TrainArgs` object containing arguments for training the model.
    """
    reset_featurization_parameters()

    if args.atom_descriptors == 'feature':
        set_extra_atom_fdim(train_args.atom_features_size)

    if args.bond_features_path is not None:
        set_extra_bond_fdim(train_args.bond_features_size)

    #set explicit H option and reaction option
    set_explicit_h(train_args.explicit_h)
    set_adding_hs(args.adding_h)
    if train_args.reaction:
        set_reaction(train_args.reaction, train_args.reaction_mode)
    elif train_args.reaction_solvent:
        set_reaction(True, train_args.reaction_mode)


def predict_and_save(args: PredictArgs, train_args: TrainArgs, test_data: MoleculeDataset,
                     task_names: List[str], num_tasks: int, test_data_loader: MoleculeDataLoader, full_data: MoleculeDataset,
                     full_to_valid_indices: dict, models: List[MoleculeModel], scalers: List[List[StandardScaler]],
                     return_invalid_smiles: bool = False):
    """
    Function to predict with a model and save the predictions to file.

    :param args: A :class:`~mixprop.args.PredictArgs` object containing arguments for
                 loading data and a model and making predictions.
    :param train_args: A :class:`~mixprop.args.TrainArgs` object containing arguments for training the model.
    :param test_data: A :class:`~mixprop.data.MoleculeDataset` containing valid datapoints.
    :param task_names: A list of task names.
    :param num_tasks: Number of tasks.
    :param test_data_loader: A :class:`~mixprop.data.MoleculeDataLoader` to load the test data.
    :param full_data:  A :class:`~mixprop.data.MoleculeDataset` containing all (valid and invalid) datapoints.
    :param full_to_valid_indices: A dictionary dictionary mapping full to valid indices.
    :param models: A list or generator object of :class:`~mixprop.models.MoleculeModel`\ s.
    :param scalers: A list or generator object of :class:`~mixprop.features.scaler.StandardScaler` objects.
    :param return_invalid_smiles: Whether to return predictions of "Invalid SMILES" for invalid SMILES, otherwise will skip them in returned predictions.
    :return:  A list of lists of target predictions.
    """
    # Predict with each model individually and sum predictions
    if args.dataset_type == 'multiclass':
        sum_preds = np.zeros((len(test_data), num_tasks, args.multiclass_num_classes))
    else:
        sum_preds = np.zeros((len(test_data), num_tasks))
    if args.ensemble_variance or args.individual_ensemble_predictions:
        if args.dataset_type == 'multiclass':
            all_preds = np.zeros((len(test_data), num_tasks, args.multiclass_num_classes, len(args.checkpoint_paths)))
        else:
            all_preds = np.zeros((len(test_data), num_tasks, len(args.checkpoint_paths)))

    # Partial results for variance robust calculation.
    print(f'Predicting with an ensemble of {len(args.checkpoint_paths)} models')
    for index, (model, scaler_list) in enumerate(tqdm(zip(models, scalers), total=len(args.checkpoint_paths))):
        scaler, features_scaler, atom_descriptor_scaler, bond_feature_scaler = scaler_list

        # Normalize features
        if args.features_scaling or train_args.atom_descriptor_scaling or train_args.bond_feature_scaling:
            test_data.reset_features_and_targets()
            if args.features_scaling:
                test_data.normalize_features(features_scaler)
            if train_args.atom_descriptor_scaling and args.atom_descriptors is not None:
                test_data.normalize_features(atom_descriptor_scaler, scale_atom_descriptors=True)
            if train_args.bond_feature_scaling and args.bond_features_size > 0:
                test_data.normalize_features(bond_feature_scaler, scale_bond_features=True)

        # Make predictions
        model_preds = predict(
            model=model,
            data_loader=test_data_loader,
            scaler=scaler
        )
        if args.dataset_type == 'spectra':
            model_preds = normalize_spectra(
                spectra=model_preds,
                phase_features=test_data.phase_features(),
                phase_mask=args.spectra_phase_mask,
                excluded_sub_value=float('nan')
            )
        sum_preds += np.array(model_preds)
        if args.ensemble_variance or args.individual_ensemble_predictions:
            if args.dataset_type == 'multiclass':
                all_preds[:,:,:,index] = model_preds
            else:
                all_preds[:,:,index] = model_preds

    # Ensemble predictions
    avg_preds = sum_preds / len(args.checkpoint_paths)

    if args.ensemble_variance:
        if args.dataset_type == 'spectra':
            all_epi_uncs = roundrobin_sid(all_preds)
        else:
            all_epi_uncs = np.var(all_preds, axis=2)
            all_epi_uncs = all_epi_uncs.tolist()

    # Save predictions
    print(f'Saving predictions to {args.preds_path}')
    #TODO: add unit tests for this
    # assert len(test_data) == len(avg_preds)
    # if args.ensemble_variance:
        # assert len(test_data) == len(all_epi_uncs)
    makedirs(args.preds_path, isfile=True)

    # Set multiclass column names, update num_tasks definition for multiclass
    if args.dataset_type == 'multiclass':
        task_names = [f'{name}_class_{i}' for name in task_names for i in range(args.multiclass_num_classes)]
        num_tasks = num_tasks * args.multiclass_num_classes

    # Copy predictions over to full_data
    for full_index, datapoint in enumerate(full_data):
        valid_index = full_to_valid_indices.get(full_index, None)
        preds = avg_preds[valid_index] if valid_index is not None else ['Invalid SMILES'] * num_tasks
        if args.ensemble_variance:
            if args.dataset_type == 'spectra':
                epi_uncs = all_epi_uncs[valid_index] if valid_index is not None else ['Invalid SMILES']
            else:
                epi_uncs = all_epi_uncs[valid_index] if valid_index is not None else ['Invalid SMILES'] * num_tasks
        if args.individual_ensemble_predictions:
            ind_preds = all_preds[valid_index] if valid_index is not None else [['Invalid SMILES'] * len(args.checkpoint_paths)] * num_tasks

        # Reshape multiclass to merge task and class dimension, with updated num_tasks
        if args.dataset_type == 'multiclass':
            if isinstance(preds, np.ndarray) and preds.ndim > 1:
                preds = preds.reshape((num_tasks))
                if args.ensemble_variance or args. individual_ensemble_predictions:
                    ind_preds = ind_preds.reshape((num_tasks, len(args.checkpoint_paths)))

        # If extra columns have been dropped, add back in SMILES columns
        if args.drop_extra_columns:
            datapoint.row = OrderedDict()

            smiles_columns = args.smiles_columns

            for column, smiles in zip(smiles_columns, datapoint.smiles):
                datapoint.row[column] = smiles

        # Add predictions columns
        for pred_name, pred in zip(task_names, preds):
            datapoint.row[pred_name] = pred
        if args.individual_ensemble_predictions:
            for pred_name, model_preds in zip(task_names,ind_preds):
                for idx, pred in enumerate(model_preds):
                    datapoint.row[pred_name+f'_model_{idx}'] = pred
        if args.ensemble_variance:
            if args.dataset_type == 'spectra':
                datapoint.row['epi_unc'] = epi_uncs
            else:
                for pred_name, epi_unc in zip(task_names, epi_uncs):
                    datapoint.row[pred_name+'_epi_unc'] = epi_unc

    # Save
    with open(args.preds_path, 'w') as f:
        writer = csv.DictWriter(f, fieldnames=full_data[0].row.keys())
        writer.writeheader()

        for datapoint in full_data:
            writer.writerow(datapoint.row)

    # Return predicted values
    avg_preds = avg_preds.tolist()
    
    if return_invalid_smiles:
        full_preds = []
        for full_index in range(len(full_data)):
            valid_index = full_to_valid_indices.get(full_index, None)
            preds = avg_preds[valid_index] if valid_index is not None else ['Invalid SMILES'] * num_tasks
            full_preds.append(preds)
        return full_preds
    else:
        return avg_preds


@timeit()
def make_predictions(args: PredictArgs, smiles: List[List[str]] = None,
                     model_objects: Tuple[PredictArgs, TrainArgs, List[MoleculeModel], List[StandardScaler], int, List[str]] = None,
                     return_invalid_smiles: bool = True,
                     return_index_dict: bool = False) -> List[List[Optional[float]]]:
    """
    Loads data and a trained model and uses the model to make predictions on the data.

    If SMILES are provided, then makes predictions on smiles.
    Otherwise makes predictions on :code:`args.test_data`.

    :param args: A :class:`~mixprop.args.PredictArgs` object containing arguments for
                 loading data and a model and making predictions.
    :param smiles: List of list of SMILES to make predictions on.
    :param model_objects: Tuple of output of load_model function which can be called separately.
    :param return_invalid_smiles: Whether to return predictions of "Invalid SMILES" for invalid SMILES, otherwise will skip them in returned predictions.
    :param return_index_dict: Whether to return the prediction results as a dictionary keyed from the initial data indexes.
    :return: A list of lists of target predictions.
    """
    if model_objects:
        args, train_args, models, scalers, num_tasks, task_names = model_objects
    else:
        args, train_args, models, scalers, num_tasks, task_names = load_model(args, generator=True)
        
    set_features(args, train_args)
    
    # Note: to get the invalid SMILES for your data, use the get_invalid_smiles_from_file or get_invalid_smiles_from_list functions from data/utils.py
    full_data, test_data, test_data_loader, full_to_valid_indices = load_data(args, smiles)
    
    # Edge case if empty list of smiles is provided
    if len(test_data) == 0:
        avg_preds = [None] * len(full_data)
    else:
        avg_preds = predict_and_save(
            args=args,
            train_args=train_args,
            test_data=test_data,
            task_names=task_names,
            num_tasks=num_tasks,
            test_data_loader=test_data_loader,
            full_data=full_data,
            full_to_valid_indices=full_to_valid_indices,
            models=models,
            scalers=scalers,
            return_invalid_smiles=return_invalid_smiles,
        )
    
    if return_index_dict:
        preds_dict = {}
        for i in range(len(full_data)):
            if return_invalid_smiles:
                preds_dict[i] = avg_preds[i]
            else:
                valid_index = full_to_valid_indices.get(i, None)
                if valid_index is not None:
                    preds_dict[i] = avg_preds[valid_index]
        return preds_dict
    else:
        return avg_preds


def mixprop_predict() -> None:
    """Parses mixprop predicting arguments and runs prediction using a trained mixprop model.

    This is the entry point for the command line command :code:`mixprop_predict`.
    """
    make_predictions(args=PredictArgs().parse_args())



import csv
from typing import List, Optional, Union

import torch
import numpy as np
from tqdm import tqdm

from mixprop.args import FingerprintArgs, TrainArgs
from mixprop.data import get_data, get_data_from_smiles, MoleculeDataLoader, MoleculeDataset
from mixprop.utils import load_args, load_checkpoint, makedirs, timeit, load_scalers, update_prediction_args
from mixprop.data import MoleculeDataLoader, MoleculeDataset
from mixprop.features import set_reaction, set_explicit_h, set_adding_hs, reset_featurization_parameters, set_extra_atom_fdim, set_extra_bond_fdim
from mixprop.models import MoleculeModel

@timeit()
def molecule_fingerprint(args: FingerprintArgs, smiles: List[List[str]] = None) -> List[List[Optional[float]]]:
    """
    Loads data and a trained model and uses the model to encode fingerprint vectors for the data.

    :param args: A :class:`~mixprop.args.PredictArgs` object containing arguments for
                 loading data and a model and making predictions.
    :param smiles: List of list of SMILES to make predictions on.
    :return: A list of fingerprint vectors (list of floats)
    """

    print('Loading training args')
    train_args = load_args(args.checkpoint_paths[0])

    # Update args with training arguments
    if args.fingerprint_type == 'MPN': # only need to supply input features if using FFN latent representation and if model calls for them.
        validate_feature_sources = False
    else:
        validate_feature_sources = True
    update_prediction_args(predict_args=args, train_args=train_args, validate_feature_sources=validate_feature_sources)
    args: Union[FingerprintArgs, TrainArgs]

    #set explicit H option and reaction option
    reset_featurization_parameters()
    if args.atom_descriptors == 'feature':
        set_extra_atom_fdim(train_args.atom_features_size)

    if args.bond_features_path is not None:
        set_extra_bond_fdim(train_args.bond_features_size)

    set_explicit_h(train_args.explicit_h)
    set_adding_hs(args.adding_h)
    if train_args.reaction:
        set_reaction(train_args.reaction, train_args.reaction_mode)
    elif train_args.reaction_solvent:
        set_reaction(True, train_args.reaction_mode)

    print('Loading data')
    if smiles is not None:
        full_data = get_data_from_smiles(
            smiles=smiles,
            skip_invalid_smiles=False,
            features_generator=args.features_generator
        )
    else:
        full_data = get_data(path=args.test_path, smiles_columns=args.smiles_columns, target_columns=[], ignore_columns=[], skip_invalid_smiles=False,
                             args=args, store_row=True)

    print('Validating SMILES')
    full_to_valid_indices = {}
    valid_index = 0
    for full_index in range(len(full_data)):
        if all(mol is not None for mol in full_data[full_index].mol):
            full_to_valid_indices[full_index] = valid_index
            valid_index += 1

    test_data = MoleculeDataset([full_data[i] for i in sorted(full_to_valid_indices.keys())])

    # Edge case if empty list of smiles is provided
    if len(test_data) == 0:
        return [None] * len(full_data)

    print(f'Test size = {len(test_data):,}')

    # Create data loader
    test_data_loader = MoleculeDataLoader(
        dataset=test_data,
        batch_size=args.batch_size,
        num_workers=args.num_workers
    )

    # Set fingerprint size
    if args.fingerprint_type == 'MPN':
        if args.atom_descriptors == "descriptor": # special case when we have 'descriptor' extra dimensions need to be added
            total_fp_size = (args.hidden_size + test_data.atom_descriptors_size()) * args.number_of_molecules
        else:
            if args.reaction_solvent:
                total_fp_size = args.hidden_size + args.hidden_size_solvent
            else:
                total_fp_size = args.hidden_size * args.number_of_molecules
        if args.features_only:
            raise ValueError('With features_only models, there is no latent MPN representation. Use last_FFN fingerprint type instead.')
    elif args.fingerprint_type == 'last_FFN':
        if args.ffn_num_layers != 1:
            total_fp_size = args.ffn_hidden_size
        else:
            raise ValueError('With a ffn_num_layers of 1, there is no latent FFN representation. Use MPN fingerprint type instead.')
    else:
        raise ValueError(f'Fingerprint type {args.fingerprint_type} not supported')
    all_fingerprints = np.zeros((len(test_data), total_fp_size*2, len(args.checkpoint_paths))) #total_fp_size, len(args.checkpoint_paths)))

    # Load model
    print(f'Encoding smiles into a fingerprint vector from {len(args.checkpoint_paths)} models.')

    for index, checkpoint_path in enumerate(tqdm(args.checkpoint_paths, total=len(args.checkpoint_paths))):
        model = load_checkpoint(checkpoint_path, device=args.device)
        scaler, features_scaler, atom_descriptor_scaler, bond_feature_scaler = load_scalers(args.checkpoint_paths[index])

        # Normalize features
        if args.features_scaling or train_args.atom_descriptor_scaling or train_args.bond_feature_scaling:
            test_data.reset_features_and_targets()
            if args.features_scaling:
                test_data.normalize_features(features_scaler)
            if train_args.atom_descriptor_scaling and args.atom_descriptors is not None:
                test_data.normalize_features(atom_descriptor_scaler, scale_atom_descriptors=True)
            if train_args.bond_feature_scaling and args.bond_features_size > 0:
                test_data.normalize_features(bond_feature_scaler, scale_bond_features=True)

        # Make fingerprints
        model_fp = model_fingerprint(
            model=model,
            data_loader=test_data_loader,
            fingerprint_type=args.fingerprint_type
        )
        if args.fingerprint_type == 'MPN' and (args.features_path is not None or args.features_generator): # truncate any features from MPN fingerprint
            model_fp = np.array(model_fp)[:,:total_fp_size] 
#         print(np.shape(all_fingerprints[:,:,index]),np.shape(model_fp))
        all_fingerprints[:,:,index] = model_fp

    # Save predictions
    print(f'Saving predictions to {args.preds_path}')
    # assert len(test_data) == len(all_fingerprints) #TODO: add unit test for this
    makedirs(args.preds_path, isfile=True)

    # Set column names
    fingerprint_columns = []
    if args.fingerprint_type == 'MPN':
        if len(args.checkpoint_paths) == 1:
            for j in range(total_fp_size//args.number_of_molecules):
                for k in range(args.number_of_molecules):
                    fingerprint_columns.append(f'fp_{j}_mol_{k}')
        else:
            for j in range(total_fp_size//args.number_of_molecules):
                for i in range(len(args.checkpoint_paths)):
                    for k in range(args.number_of_molecules):
                        fingerprint_columns.append(f'fp_{j}_mol_{k}_model_{i}')

    else: # args == 'last_FNN'
        if len(args.checkpoint_paths) == 1:
            for j in range(total_fp_size):
                fingerprint_columns.append(f'fp_{j}')
        else:
            for j in range(total_fp_size):
                for i in range(len(args.checkpoint_paths)):
                    fingerprint_columns.append(f'fp_{j}_model_{i}')

    # Copy predictions over to full_data
    for full_index, datapoint in enumerate(full_data):
        valid_index = full_to_valid_indices.get(full_index, None)
        preds = all_fingerprints[valid_index].reshape((len(args.checkpoint_paths) * total_fp_size*2)) if valid_index is not None else ['Invalid SMILES'] * len(args.checkpoint_paths) * total_fp_size*2

        for i in range(len(fingerprint_columns)):
            datapoint.row[fingerprint_columns[i]] = preds[i]

    # Write predictions
    with open(args.preds_path, 'w') as f:
        writer = csv.DictWriter(f, fieldnames=args.smiles_columns+fingerprint_columns,extrasaction='ignore')
        writer.writeheader()
        for datapoint in full_data:
            writer.writerow(datapoint.row)

    return all_fingerprints

def model_fingerprint(model: MoleculeModel,
            data_loader: MoleculeDataLoader,
            fingerprint_type: str = 'MPN',
            disable_progress_bar: bool = False) -> List[List[float]]:
    """
    Encodes the provided molecules into the latent fingerprint vectors, according to the provided model.

    :param model: A :class:`~mixprop.models.model.MoleculeModel`.
    :param data_loader: A :class:`~mixprop.data.data.MoleculeDataLoader`.
    :param disable_progress_bar: Whether to disable the progress bar.
    :return: A list of fingerprint vector lists.
    """
    model.eval()

    fingerprints = []

    for batch in tqdm(data_loader, disable=disable_progress_bar, leave=False):
        # Prepare batch
        batch: MoleculeDataset
        mol_batch, features_batch, atom_descriptors_batch, atom_features_batch, bond_features_batch = \
            batch.batch_graph(), batch.features(), batch.atom_descriptors(), batch.atom_features(), batch.bond_features()

        # Make predictions
        with torch.no_grad():
            batch_fp = model.fingerprint(mol_batch, features_batch, atom_descriptors_batch,
                                atom_features_batch, bond_features_batch, fingerprint_type)

        # Collect vectors
        batch_fp = batch_fp.data.cpu().tolist()

        fingerprints.extend(batch_fp)

    return fingerprints

def mixprop_fingerprint() -> None:
    """
    Parses mixprop predicting arguments and returns the latent representation vectors for
    provided molecules, according to a previously trained model.
    """
    molecule_fingerprint(args=FingerprintArgs().parse_args())
